################################################################################
# COPYRIGHT(c) 2018 STMicroelectronics                                         #
#                                                                              #
# Redistribution and use in source and binary forms, with or without           #
# modification, are permitted provided that the following conditions are met:  #
#   1. Redistributions of source code must retain the above copyright notice,  #
#      this list of conditions and the following disclaimer.                   #
#   2. Redistributions in binary form must reproduce the above copyright       #
#      notice, this list of conditions and the following disclaimer in the     #
#      documentation and/or other materials provided with the distribution.    #
#   3. Neither the name of STMicroelectronics nor the names of its             #
#      contributors may be used to endorse or promote products derived from    #
#      this software without specific prior written permission.                #
#                                                                              #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  #
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    #
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   #
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE    #
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR          #
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF         #
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS     #
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN      #
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)      #
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE   #
# POSSIBILITY OF SUCH DAMAGE.                                                  #
################################################################################


"""iolink_master

The iolink_master module is responsible for managing IO-Link capable devices
connected to an iolink masterboard and allocating the needed resources.
"""


# IMPORT

from abc import ABCMeta
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from serial import SerialException
from serial import SerialTimeoutException
import time
from datetime import datetime

from wire_st_sdk.iolink.iolink_protocol import IOLinkProtocol
from wire_st_sdk.iolink.iolink_sensor import IOLinkSensor
from wire_st_sdk.utils.wire_st_exceptions import InvalidOperationException
from wire_st_sdk.python_utils import lock
from wire_st_sdk.python_utils import lock_for_object


# CLASSES

class IOLinkMaster(object):
    """IO-Link Master class.

    This class manages the communication between the host and the masterboard.
    """

    _NUMBER_OF_DEVICES = 4
    """Number of devices that can be connected to the masterboard."""

    _CONFIGURATION_REGISTERS_STRING = '096,248,033,122,122,122,122,' \
        + IOLinkProtocol.TERMINATOR_SEQ
    """Configuration string for the masterboard's registers."""

    _NUMBER_OF_THREADS = 5
    """Number of threads to be used to notify the listeners."""

    _SERIAL_PORT_RW_DELAY_s = 0.1
    """Delay time in seconds between two consecutive reads/writes on the serial
    port."""

    _FDM_ATTEMPTS = 10
    """Number of attempts to receive a line of frequency domain data."""

    def __init__(self, serial_port):
        """Constructor.

        Args:
            serial_port (Serial): Serial Port object. Refer to
                `Serial <https://pyserial.readthedocs.io/en/latest/pyserial_api.html#serial.Serial>`_
                for more information.
        """
    
        self._serial_port = serial_port
        """Serial Port object."""

        self._answer = None
        """Last answer received on the serial port when executing a command."""

        self._status = IOLinkMasterStatus.INIT
        """Status."""

        self._thread_pool = ThreadPoolExecutor(IOLinkMaster._NUMBER_OF_THREADS)
        """Pool of thread used to notify the listeners."""

        self._listeners = []
        """List of listeners to the master changes.
        It is a thread safe list, so a listener can subscribe itself through a
        callback."""

        self._devices = [None] * self._NUMBER_OF_DEVICES
        """List of devices by position."""

        self._devices_id_position_map = {}
        """Dictionary that maps devices' identifier and devices' position."""

        # Updating status.
        self._update_master_status(IOLinkMasterStatus.IDLE)

    def _update_master_status(self, new_status):
        """Update the status of the master.

        Args:
            new_status (:class:`wire_st_sdk.iolink.master.IOLinkMasterStatus`):
                New status.
        """
        old_status = self._status
        self._status = new_status
        for listener in self._listeners:
            # Calling user-defined callback.
            self._thread_pool.submit(
                listener.on_status_change(
                    self, new_status.value, old_status.value))

    def _update_master_device(self, device_id, device_position):
        """Update about the new device found.

        Args:
            device_id (str): New device found.
            device_position (int): Position of the new device found.
        """
        for listener in self._listeners:
            # Calling user-defined callback.
            self._thread_pool.submit(
                listener.on_device_found(
                    self, device_id, device_position))

    def _execute(self, command=None, expected_answer=None):
        """Execute a command and check the status.

        Args:
            command (str): Command to perform.
            expected_answer (str): Expected answer to check after performing the
                command.

        Returns:
            bool: True if the command has been executed correctly, False
            otherwise.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self):

                #print('---> %s %s' % ((command, expected_answer)))

                # Delay.
                time.sleep(self._SERIAL_PORT_RW_DELAY_s)

                # Blocking write.
                if command:
                    #print('---> \"%s\"' % (command[:-2]))
                    self._serial_port.write(command)
                    self._serial_port.flush()

                # Delay.
                time.sleep(self._SERIAL_PORT_RW_DELAY_s)

                # Blocking read.
                if expected_answer:
                    if not (command == IOLinkProtocol.COMMAND_MEAS1_4 and \
                        IOLinkProtocol.BYTES_TRANSMISSION):
                        self._answer = ''
                        while True:
                            self._answer += \
                                self._serial_port.read_until(expected_answer)
                            if expected_answer in self._answer:
                                return True
                            if IOLinkProtocol.MESSAGE_SENSOR_FAILED in \
                                self._answer or \
                                IOLinkProtocol.MESSAGE_PARAMETER_UPDATED in \
                                self._answer:
                                return False
                    else:
                        USE_ACK = False
                        self._answer = ''
                        n = IOLinkSensor._SIZE_OF_FDM * \
                            IOLinkSensor._SIZE_OF_FLOAT_bytes
                        attempts = self._FDM_ATTEMPTS
                        while attempts:
                            data = self._serial_port.read(n)
                            if len(data) == n:
                                attempts = self._FDM_ATTEMPTS
                                self._answer += data
                                if USE_ACK:
                                    self._serial_port.write(IOLinkProtocol.ACK)
                                    self._serial_port.flush()
                            else:
                                if USE_ACK:
                                    self._serial_port.write(IOLinkProtocol.NACK)
                                    self._serial_port.flush()
                                    attempts -= 1
                                    print('Trying again...(%d)' % (attempts))
                                else:
                                    return False
                            if len(self._answer) == 1024 * n:
                                while True:
                                    self._answer += \
                                        self._serial_port.read_until(
                                            expected_answer)
                                    if expected_answer in self._answer:
                                        return True
                        return False

                # Blocking read.
                # self._answer = ''
                # if expected_answer:
                #     while expected_answer not in self._answer:
                #         self._answer += self._serial_port.read_until(expected_answer)
                #         if IOLinkProtocol.MESSAGE_SENSOR_FAILED in self._answer or \
                #             IOLinkProtocol.MESSAGE_PARAMETER_UPDATED in self._answer:
                #             # Sensor has already been rebooted.
                #             return False

                # # Non-blocking read.
                # self._answer = self._serial_port.read(1)
                # while expected_answer not in self._answer:
                #     while (self._serial_port.in_waiting > 0):
                #         self._answer += self._serial_port.read(self._serial_port.in_waiting)
                #     if IOLinkProtocol.MESSAGE_SENSOR_FAILED in self._answer or \
                #              IOLinkProtocol.MESSAGE_PARAMETER_UPDATED in self._answer:
                #              return False

                return True

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def connect(self):
        """Start the communication between the masterboard and the host to which
        it is connected.
        
        Returns:
            str: The status of the connection.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:

            with lock_for_object(self):

                # Updating status.
                self._update_master_status(IOLinkMasterStatus.CONNECTING)

                # Opening the serial port.

                if not self._serial_port.is_open:
                    self._serial_port.open()

                # Configuring all devices.

                self._execute(IOLinkProtocol.COMMAND_START,
                    IOLinkProtocol.REQUEST_MOD)

                connection_answer = "\r\n" + self._answer[: \
                    - len(IOLinkProtocol.REQUEST_MOD) \
                    - len(IOLinkProtocol.TERMINATOR_SEQ)]

                for device_position in range(self._NUMBER_OF_DEVICES):

                    self._execute(IOLinkProtocol.COMMAND_ICM,
                        IOLinkProtocol.REQUEST_IC)

                    self._execute(str(device_position) \
                        + IOLinkProtocol.TERMINATOR_SEQ,
                        IOLinkProtocol.REQUEST_OPM)

                    self._execute(IOLinkProtocol.COMMAND_WRS,
                        IOLinkProtocol.REQUEST_RVAL)

                    self._execute(self._CONFIGURATION_REGISTERS_STRING,
                        IOLinkProtocol.MESSAGE_PROGRAMMING_DONE)

                    self._execute(IOLinkProtocol.COMMAND_END,
                        IOLinkProtocol.REQUEST_MOD)

                # Clearing devices' maps.

                self._devices = [None] * self._NUMBER_OF_DEVICES
                self._devices_id_position_map.clear()

                # Getting all devices' identifiers.

                for device_position in range(self._NUMBER_OF_DEVICES):

                    self._execute(IOLinkProtocol.COMMAND_ICD,
                        IOLinkProtocol.REQUEST_SLAVE)

                    self._execute(str(device_position) \
                        + IOLinkProtocol.TERMINATOR_SEQ,
                        IOLinkProtocol.REQUEST_SENSOR_COMMAND)

                    self._timeout = self._serial_port.timeout
                    self._serial_port.timeout = 5

                    self._execute(IOLinkProtocol.COMMAND_GET_1,
                        IOLinkProtocol.TERMINATOR_SEQ)

                    self._serial_port.timeout = self._timeout

                    self._execute(None, IOLinkProtocol.TERMINATOR_SEQ)

                    # Handling sensor error.
                    if IOLinkProtocol.MESSAGE_SENSOR_FAILED in self._answer or \
                        IOLinkProtocol.MESSAGE_PARAMETER_UPDATED in self._answer:
                        # Sensor has already been rebooted.
                        #print('Rebooting sensor...')
                        #print('---> "MCU" KO: \"%s\"' % (self._answer[:-2]))
                        self._execute(IOLinkProtocol.COMMAND_END,
                            IOLinkProtocol.REQUEST_MOD)
                        continue

                    # Getting device's identifier, build devices' map, and
                    # waiting for transmission completed.
                    device_id = self.get_answer()[:
                        - len(IOLinkProtocol.TERMINATOR_SEQ)].strip()

                    self._devices[device_position] = device_id

                    self._devices_id_position_map[device_id] = \
                        device_position + 1

                    self._execute(None, IOLinkProtocol.TERMINATOR_SEQ)

                    self._execute(None,
                        IOLinkProtocol.MESSAGE_TRANSMISSION_COMPLETED)

                    self._execute(IOLinkProtocol.COMMAND_END,
                        IOLinkProtocol.REQUEST_MOD)

                    # Updating about new device found.
                    self._update_master_device(device_id, device_position + 1)

                self._answer = connection_answer

                # Updating status.
                self._update_master_status(IOLinkMasterStatus.CONNECTED)

                return self._answer

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def disconnect(self):
        """End the communication between the masterboard and the host to which
        it is connected.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
        """
        try:
            if not self.is_connected():
                return

            with lock_for_object(self):

                # Updating status.
                self._update_master_status(IOLinkMasterStatus.DISCONNECTING)

                # Clearing devices' maps.
                self._devices = [None] * self._NUMBER_OF_DEVICES
                self._devices_id_position_map.clear()

                # Closing the serial port.
                if self._serial_port.is_open:
                    self._serial_port.close()

                # Updating status.
                self._update_master_status(IOLinkMasterStatus.IDLE)

        except (SerialException, SerialTimeoutException) as e:
            raise e

    def is_connected(self):
        """Check whether the master is connected.

        Returns:
            bool: True if the master is connected, False otherwise.
        """
        return self._answer == IOLinkMasterStatus.CONNECTED

    def get_device(self, device_id, device_name=None):
        """Configure and get a connected device.

        Args:
            device_id (str): Device's identifier.
            device_name (str): Device's name.

        Returns:
            :class:`wire_st_sdk.iolink.iolink_device.IOLinkDevice`: IOLink
            connected device if the device identifier is correct, None
            otherwise.

        Raises:
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        # Creating the device.
        try:
            with lock_for_object(self):
                if self.get_answer() == None:
                    raise InvalidOperationException('Before trying to get a \
                        device, connect to the masterboard.')

                if device_id in self._devices_id_position_map:
                    device_position = self._devices_id_position_map[device_id]
                    return IOLinkSensor(
                        self, device_position, device_id, device_name)

                return None

        except InvalidOperationException as e:
            raise e

    def get_device_by_position(self, device_position, device_name=None):
        """Configure and get a connected device.

        Args:
            device_position (int): Device's position according to the
                enumeration on the masterboard.
            device_name (str): Device's name.

        Returns:
            :class:`wire_st_sdk.iolink.iolink_device.IOLinkDevice`: IOLink
            connected device if there is a device connected at the given device
            position, None otherwise.

        Raises:
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
            :exc:`ValueError` if the device position is not allowed.
        """
        # Creating the device.
        try:
            with lock_for_object(self):
                if self.get_answer() == None:
                    raise InvalidOperationException('Before trying to get a \
                        device, connect to the masterboard.')

                if device_position not in range(1, self._NUMBER_OF_DEVICES + 1):
                    raise ValueError('Device\'s position must be in range ' \
                        '[%d..%d].' % (1, self._NUMBER_OF_DEVICES))

                if device_position in range(1, self._NUMBER_OF_DEVICES + 1):
                    device_id = self._devices[device_position - 1]
                    return IOLinkSensor(
                        self, device_position, device_id, device_name)

                return None

        except InvalidOperationException as e:
            raise e

    def get_answer(self):
        """Get the last answer received on the serial port when executing a
        command."""
        return self._answer

    def get_port(self):
        """Get the serial port of the masterboard.

        Returns:
            Serial: Serial Port object. Refer to
                `Serial <https://pyserial.readthedocs.io/en/latest/pyserial_api.html#serial.Serial>`_
                for more information.
        """
        return self._serial_port

    def add_listener(self, listener):
        """Add a listener.
        
        Args:
            listener (:class:`wire_st_sdk.iolink.MasterListener`): Listener to
                be added.
        """
        if listener is not None:
            with lock(self):
                if not listener in self._listeners:
                    self._listeners.append(listener)

    def remove_listener(self, listener):
        """Remove a listener.

        Args:
            listener (:class:`wire_st_sdk.iolink.MasterListener`): Listener to
                be removed.
        """
        if listener is not None:
            with lock(self):
                if listener in self._listeners:
                    self._listeners.remove(listener)


class IOLinkMasterStatus(Enum):
    """Status of the masterboard."""

    INIT = 'INIT'
    """Dummy initial status."""

    IDLE = 'IDLE'
    """Waiting for a connection."""

    CONNECTING = 'CONNECTING'
    """Opening a connection with the host."""

    CONNECTED = 'CONNECTED'
    """Connected to the host."""

    DISCONNECTING = 'DISCONNECTING'
    """Closing the connection to the host."""


# INTERFACES

class IOLinkMasterListener(object):
    """Interface used by the :class:`wire_st_sdk.iolink.IOLinkMaster` class to
    notify changes of a masterboard's status.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def on_status_change(self, masterboard, new_status, old_status):
        """To be called whenever a masterboard changes its status.

        Args:
            masterboard (IOLinkMaster): Masterboard that has changed its status.
            new_status (:class:`wire_st_sdk.iolink.IOLinkMasterStatus`): New
                status.
            old_status (:class:`wire_st_sdk.iolink.IOLinkMasterStatus`): Old
                status.

        Raises:
            'NotImplementedError' is raised if the method is not implemented.
        """
        raise NotImplementedError('You must define "on_status_change()" to use '
            'the "IOLinkMasterListener" class.')

    @abstractmethod
    def on_device_found(self, masterboard, device_id, device_position):
        """To be called whenever a masterboard finds a new device connected.

        Args:
            masterboard (IOLinkMaster): Masterboard that has found a new device.
            device_id (str): New device found.
            device_position (int): Position of the new device found.

        Raises:
            'NotImplementedError' is raised if the method is not implemented.
        """
        raise NotImplementedError('You must define "on_device_found()" to use '
            'the "IOLinkMasterListener" class.')



################################################################################
# COPYRIGHT(c) 2018 STMicroelectronics                                         #
#                                                                              #
# Redistribution and use in source and binary forms, with or without           #
# modification, are permitted provided that the following conditions are met:  #
#   1. Redistributions of source code must retain the above copyright notice,  #
#      this list of conditions and the following disclaimer.                   #
#   2. Redistributions in binary form must reproduce the above copyright       #
#      notice, this list of conditions and the following disclaimer in the     #
#      documentation and/or other materials provided with the distribution.    #
#   3. Neither the name of STMicroelectronics nor the names of its             #
#      contributors may be used to endorse or promote products derived from    #
#      this software without specific prior written permission.                #
#                                                                              #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  #
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    #
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   #
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE    #
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR          #
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF         #
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS     #
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN      #
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)      #
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE   #
# POSSIBILITY OF SUCH DAMAGE.                                                  #
################################################################################


"""iolink_protocol

The iolink_protocol module contains definitions related to the IO-Link protocol.
"""


# IMPORT

from enum import Enum


# COMMAND STRINGS

class IOLinkProtocol(object):
    """This class lists the protocol's commands used for IO-Link communication.
    """

    # Data transmission.

    BYTES_TRANSMISSION = True
    """If True, transmission of data is by bytes; by characters otherwise."""


    # Common sequences.

    TERMINATOR_SEQ = '\r\n'
    """Terminator sequence."""
    ACK = '1'
    """Acknowledgement."""
    NACK = '0'
    """Non-Acknowledgement."""


    # Commands to the masterboard.

    COMMAND_START = 'START' + TERMINATOR_SEQ
    """Open RS485 communication."""
    COMMAND_WRC = 'WR_C' + TERMINATOR_SEQ
    """L6360 write curent mode."""
    COMMAND_RDC = 'RD_C' + TERMINATOR_SEQ
    """L6360 read curent mode."""
    COMMAND_WRS = 'WR_S' + TERMINATOR_SEQ
    """L6360 write sequential mode."""
    COMMAND_RDS = 'RD_S' + TERMINATOR_SEQ
    """L6360 read sequential mode."""
    COMMAND_ICM = 'MASTER' + TERMINATOR_SEQ
    """Master settings mode."""
    COMMAND_ICD = 'DEVICE' + TERMINATOR_SEQ
    """IO-Link communication mode."""
    COMMAND_NEW = 'COMMAND NEW' + TERMINATOR_SEQ
    """Continue programming on the same node."""
    COMMAND_END = 'COMMAND END' + TERMINATOR_SEQ
    """Close actual programming mode and change node."""
    COMMAND_ID = 'IDS' + TERMINATOR_SEQ
    """Get sensors list."""
    COMMAND_SP = 'PRM' + TERMINATOR_SEQ
    """Get sensor's parameter."""
    COMMAND_SP1 = 'SENSOR ACC PRM' + TERMINATOR_SEQ
    """Get accelerometer's data."""
    COMMAND_SP2 = 'SENSOR ENV PRM' + TERMINATOR_SEQ
    """Get evironmental data."""
    COMMAND_MEAS1 = 'MSR' + TERMINATOR_SEQ
    """Request for measurement."""
    COMMAND_MEAS1_1 = 'RAW' + TERMINATOR_SEQ
    """RAW data."""
    COMMAND_MEAS1_2 = 'TDM' + TERMINATOR_SEQ
    """Time domain data."""
    COMMAND_MEAS1_3 = 'ENV' + TERMINATOR_SEQ
    """Environmental data."""
    COMMAND_MEAS1_4 = 'FFT' + TERMINATOR_SEQ
    """Fast Fourier Transform data."""
    COMMAND_GET = 'STS' + TERMINATOR_SEQ
    """Get sensor's status, vibration peak, RMS, or vibration overtoken."""
    COMMAND_GET_1 = 'MCU' + TERMINATOR_SEQ
    """Microcontroller's identifier."""
    COMMAND_SET1 = 'SENSOR CNF' + TERMINATOR_SEQ
    """Sensor configuration parameters."""
    COMMAND_SET1_1 = 'SENSOR ACC CNF' + TERMINATOR_SEQ
    """Sensor configuration accelerometer parameters."""
    COMMAND_SET1_2 = 'SENSOR ACC CPT' + TERMINATOR_SEQ
    """Sensor configuration computation parameters."""
    COMMAND_SET = 'SET' + TERMINATOR_SEQ
    """Set sensor parameter."""
    COMMAND_ODR = 'ODR' + TERMINATOR_SEQ
    """Set accelerometer's output data rate."""
    COMMAND_FLS = 'FLS' + TERMINATOR_SEQ
    """Set accelerometer's full scale."""
    COMMAND_SZE = 'SZE' + TERMINATOR_SEQ
    """Set accelerometer's input array size for FFT."""
    COMMAND_SUB = 'SUB' + TERMINATOR_SEQ
    """Set accelerometer's number of subranges."""
    COMMAND_ACQ = 'ACQ' + TERMINATOR_SEQ
    """Set accelerometer's total acquisition time (all types of analysis)."""
    COMMAND_OVL = 'OVL' + TERMINATOR_SEQ
    """Set accelerometer's overlapping percentage between two consecutive FFT
    analysis."""


    # Requests from the masterboard.

    REQUEST_MOD = 'SELECT MASTER OR DEVICE' + TERMINATOR_SEQ
    REQUEST_IC = 'INSERT ADDRESS IC :' + TERMINATOR_SEQ
    REQUEST_OPM = 'INSERT OPERATING MODE :' + TERMINATOR_SEQ
    REQUEST_RADRS = 'INSERT REGISTER ADDRESS :' + TERMINATOR_SEQ
    REQUEST_RVAL = 'INSERT REGISTER VALUE :' + TERMINATOR_SEQ
    REQUEST_PARAMETER_NAME = 'INSERT PARAMETER NAME :'
    REQUEST_PARAMETER_VALUE = 'INSERT PARAMETER VALUE :'
    REQUEST_SLAVE = 'INSERT SLAVE NODE :'
    REQUEST_SENSOR_COMMAND = 'INSERT SENSOR COMMAND :'
    REQUEST_MEASURE_TYPE = 'INSERT MEASURE TYPE :'


    # Messages from the masterboard.

    MESSAGE_PROGRAMMING_DONE = 'PROGRAMMING DONE' + TERMINATOR_SEQ \
        + TERMINATOR_SEQ
    MESSAGE_CONNECTED = 'CONNECTED' + TERMINATOR_SEQ
    MESSAGE_TRANSMISSION_COMPLETED = 'TRANSMISSION COMPLETED' + TERMINATOR_SEQ
    MESSAGE_SENSOR_FAILED = 'SENSOR FAILED' + TERMINATOR_SEQ
    MESSAGE_PARAMETER_UPDATED = 'PARAMETER UPDATED' + TERMINATOR_SEQ
    MESSAGE_PARAMETER_ERROR = 'PARAMETER ERROR' + TERMINATOR_SEQ


# PARAMETER VALUES

class ODR(Enum):
    """Allowed parameters' values for ODR.
    Accelerometer's output data rate."""
    ODR_6660 = 6660
    ODR_3330 = 3330
    ODR_1660 = 1660
    ODR_833 = 833
    ODR_416 = 416
    ODR_208 = 208
    ODR_104 = 104
    ODR_52 = 52
    ODR_26 = 26
    ODR_13 = 13


class FLS(Enum):
    """Allowed parameters' values for FLS.
    Accelerometer's full scale."""
    FLS_2 = 2
    FLS_4 = 4
    FLS_8 = 8
    FLS_16 = 16


class SZE(Enum):
    """Allowed parameters' values for SZE.
    Accelerometer's input array size for FFT."""
    SZE_256 = 256
    SZE_512 = 512
    SZE_1024 = 1024
    SZE_2048 = 2048


class SUB(Enum):
    """Allowed parameters' values for SUB.
    Accelerometer's number of subranges."""
    SUB_8 = 8
    SUB_16 = 16
    SUB_32 = 32
    SUB_64 = 64


# Allowed parameters' values for ACQ. 
# Accelerometer's total acquisition time.
ACQ_MIN = 500
ACQ_MAX = 60000


# Allowed parameters' values for OVL.
# Accelerometer's overlapping percentage between two consecutive FFT analysis.
OVL_MIN = 5
OVL_MAX = 95



################################################################################
# COPYRIGHT(c) 2018 STMicroelectronics                                         #
#                                                                              #
# Redistribution and use in source and binary forms, with or without           #
# modification, are permitted provided that the following conditions are met:  #
#   1. Redistributions of source code must retain the above copyright notice,  #
#      this list of conditions and the following disclaimer.                   #
#   2. Redistributions in binary form must reproduce the above copyright       #
#      notice, this list of conditions and the following disclaimer in the     #
#      documentation and/or other materials provided with the distribution.    #
#   3. Neither the name of STMicroelectronics nor the names of its             #
#      contributors may be used to endorse or promote products derived from    #
#      this software without specific prior written permission.                #
#                                                                              #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  #
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    #
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   #
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE    #
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR          #
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF         #
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS     #
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN      #
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)      #
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE   #
# POSSIBILITY OF SUCH DAMAGE.                                                  #
################################################################################


"""iolink_device

The iolink_device module represent a device capable of connecting to an IO-Link
masterboard and sending/receiving data to it.
"""


# IMPORT

from serial import SerialException
from serial import SerialTimeoutException

from wire_st_sdk.iolink.iolink_protocol import IOLinkProtocol
from wire_st_sdk.utils.wire_st_exceptions import InvalidOperationException
from wire_st_sdk.python_utils import lock_for_object


# CLASSES

class IOLinkDevice(object):
    """IO-Link Device class.

    This class manages the commands of a standard IO-Link device.
    """

    def __init__(self, master, position, id, name=None):
        """Constructor.

        Args:
            master (:class:`wire_st_sdk.iolink.iolink_master.IOLinkMaster`):
                Masterboard object.
            position (int): Device's position according to the enumeration on
                the masterboard.
            id (str): Device's identifier.
            name (str): Device's name.
        """

        self._master = master
        """Masterboard object."""

        self._position = position
        """Device's position according to the enumeration on the masterboard."""

        self._id = id
        """Device's identifier."""

        self._name = name
        """Device's name."""

    def get_position(self):
        """Get the device's position according to the enumeration on the
        masterboard.

        Returns:
            int: The device's position according to the enumeration on the
            masterboard.
        """
        return self._position

    def get_id(self):
        """Get the device's identifier.

        Returns:
            str: The device's identifier.
        """
        return self._id

    def get_name(self):
        """Get the device's name.

        Returns:
            str: The device's name.
        """
        return self._name

    def get_features(self):
        """Get the list of features.

        Returns:
            list: A list with the available features.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                self._master._execute(IOLinkProtocol.COMMAND_ICD,
                    IOLinkProtocol.REQUEST_SLAVE)

                self._master._execute(str(self._position - 1) \
                    + IOLinkProtocol.TERMINATOR_SEQ,
                    IOLinkProtocol.REQUEST_SENSOR_COMMAND)

                self._master._execute(IOLinkProtocol.COMMAND_ID,
                    IOLinkProtocol.MESSAGE_CONNECTED)
                
                info = self._master.get_answer()[
                    len(IOLinkProtocol.TERMINATOR_SEQ): \
                    - len(IOLinkProtocol.MESSAGE_CONNECTED) \
                    - 2 * len(IOLinkProtocol.TERMINATOR_SEQ)]
                info = info.split(IOLinkProtocol.TERMINATOR_SEQ)
                info = list(map(lambda s: s.strip(), info))[:-1]

                self._master._execute(IOLinkProtocol.COMMAND_END,
                    IOLinkProtocol.REQUEST_MOD)

                return info

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def get_firmware(self):
        """Get the version of the firmware.

        Returns:
            str: The version of the firmware as a string.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                self._master._execute(IOLinkProtocol.COMMAND_ICD,
                    IOLinkProtocol.REQUEST_SLAVE)

                self._master._execute(str(self._position - 1) \
                    + IOLinkProtocol.TERMINATOR_SEQ,
                    IOLinkProtocol.REQUEST_SENSOR_COMMAND)

                self._master._execute(IOLinkProtocol.COMMAND_ID,
                    IOLinkProtocol.MESSAGE_CONNECTED)
                
                info = self._master.get_answer()[
                    len(IOLinkProtocol.TERMINATOR_SEQ): \
                    - len(IOLinkProtocol.MESSAGE_CONNECTED) \
                    - 2 * len(IOLinkProtocol.TERMINATOR_SEQ)]
                info = info.split(IOLinkProtocol.TERMINATOR_SEQ)[-1]

                self._master._execute(IOLinkProtocol.COMMAND_END,
                    IOLinkProtocol.REQUEST_MOD)

                return info

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e



################################################################################
# COPYRIGHT(c) 2018 STMicroelectronics                                         #
#                                                                              #
# Redistribution and use in source and binary forms, with or without           #
# modification, are permitted provided that the following conditions are met:  #
#   1. Redistributions of source code must retain the above copyright notice,  #
#      this list of conditions and the following disclaimer.                   #
#   2. Redistributions in binary form must reproduce the above copyright       #
#      notice, this list of conditions and the following disclaimer in the     #
#      documentation and/or other materials provided with the distribution.    #
#   3. Neither the name of STMicroelectronics nor the names of its             #
#      contributors may be used to endorse or promote products derived from    #
#      this software without specific prior written permission.                #
#                                                                              #
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"  #
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE    #
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE   #
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE    #
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR          #
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF         #
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS     #
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN      #
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)      #
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE   #
# POSSIBILITY OF SUCH DAMAGE.                                                  #
################################################################################


"""iolink_sensor

The iolink_sensor module represents a device capable of connecting to an IOLink
masterboard and sending/receiving data to it.
"""


# IMPORT

import sys
import struct

from serial import SerialException
from serial import SerialTimeoutException

from wire_st_sdk.iolink.iolink_protocol import IOLinkProtocol
from wire_st_sdk.iolink.iolink_device import IOLinkDevice
from wire_st_sdk.utils.wire_st_exceptions import InvalidOperationException
from wire_st_sdk.python_utils import lock_for_object


# CLASSES

class IOLinkSensor(IOLinkDevice):
    """IO-Link Sensor class.

    This class manages the commands of a standard IO-Link sensor.
    """

    _SIZE_OF_ENV = 3
    """Number of elements in an environmental domain data."""

    _SIZE_OF_TDM = 6
    """Number of elements in a time domain data."""

    _SIZE_OF_FDM = 4
    """Number of elements per frequency in a frequency domain data."""

    _SIZE_OF_FDM_LINES = 1024
    """Number of elements per frequency in a frequency domain data."""

    _SIZE_OF_FLOAT_bytes = 4
    """Size of floating point numbers in bytes on ARM-32 platforms."""

    _FLOAT_PRECISION = 3
    """Number of digits after the decimal point for sensors' data."""

    def _get_measure(self, measure):
        """Execute a 'get' measure command.

        Args:
            measure (str): The measure command to execute.

        Returns:
            list: A list with the result of the executed measure command.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                while True:
                    self._master._execute(IOLinkProtocol.COMMAND_ICD,
                        IOLinkProtocol.REQUEST_SLAVE)

                    self._master._execute(str(self._position - 1) \
                        + IOLinkProtocol.TERMINATOR_SEQ,
                        IOLinkProtocol.REQUEST_SENSOR_COMMAND)

                    self._master._execute(IOLinkProtocol.COMMAND_MEAS1,
                        IOLinkProtocol.REQUEST_MEASURE_TYPE)

                    if not self._master._execute(measure,
                        IOLinkProtocol.MESSAGE_TRANSMISSION_COMPLETED):

                        self._master._execute(IOLinkProtocol.COMMAND_END,
                        IOLinkProtocol.REQUEST_MOD)
                    else:
                        #print('TRANSMISSION COMPLETED')
                        break

                if not IOLinkProtocol.BYTES_TRANSMISSION:
                    info = self._master.get_answer()[
                        len(IOLinkProtocol.TERMINATOR_SEQ): \
                        - 2 * len(IOLinkProtocol.TERMINATOR_SEQ)
                        - len(IOLinkProtocol.MESSAGE_TRANSMISSION_COMPLETED)]
                else:
                    info = self._master.get_answer()[: \
                        - len(IOLinkProtocol.TERMINATOR_SEQ)
                        - len(IOLinkProtocol.MESSAGE_TRANSMISSION_COMPLETED)]

                #print('SENDING COMMAND END')
                self._master._execute(IOLinkProtocol.COMMAND_END,
                    IOLinkProtocol.REQUEST_MOD)
                #print('COMMAND END SENT')

                return info

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def _bytes_to_floats(self, data, precision=0):
        """Converts an array of bytes to floating point numbers in Little Endian
        order (four bytes per number).

        Args:
            data (str): Input array of bytes that contains the values to convert.
            precision (int): Number of digits after the decimal point.

        Returns:
            list: A list of floating point numbers.
        """
        return [round(
            struct.unpack('<f', struct.pack('cccc', *data[i * 4:i * 4 + 4]))[0],
            precision) \
            for i in range(0, len(data) / 4)]

    def get_env(self):
        """Get environmental data.

        Returns:
            list: A list with Pressure [mbar], Humidity [%], and Temperature [C]
            values.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                while True:
                    info = self._get_measure(IOLinkProtocol.COMMAND_MEAS1_3)

                    if not IOLinkProtocol.BYTES_TRANSMISSION:
                        info = info.split(IOLinkProtocol.TERMINATOR_SEQ)[2:]
                        info = [i for i in info if i != '']
                        if len(info) == IOLinkSensor._SIZE_OF_ENV:
                            break
                    else:
                        if len(info) == IOLinkSensor._SIZE_OF_ENV * \
                            IOLinkSensor._SIZE_OF_FLOAT_bytes:
                            break

                    print('Missing data. Sending again....')

                if not IOLinkProtocol.BYTES_TRANSMISSION:
                    info = list(map(lambda s: float(s), info))
                else:
                    info = self._bytes_to_floats(info, self._FLOAT_PRECISION)

                return info

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def get_tdm(self):
        """Get time domain data.

        Returns:
            list: A two-elements list, with a list of RMS Speed values on X,Y,Z
            axes [mm/s] as the first element, and a list of Peak Acceleration
            values on X,Y,Z axes [m/s2] as the second element.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                while True:
                    info = self._get_measure(IOLinkProtocol.COMMAND_MEAS1_2)

                    if not IOLinkProtocol.BYTES_TRANSMISSION:
                        info = info.split(IOLinkProtocol.TERMINATOR_SEQ)[2:]
                        if len(info) == IOLinkSensor._SIZE_OF_TDM:
                            break
                    else:
                        if len(info) == IOLinkSensor._SIZE_OF_TDM * \
                            IOLinkSensor._SIZE_OF_FLOAT_bytes:
                            break

                    print('Missing data. Sending again....')

                if not IOLinkProtocol.BYTES_TRANSMISSION:
                    info = [list(map(lambda s: float(s), info[0:3])), \
                            list(map(lambda s: float(s), info[3:6]))]
                else:
                    info = self._bytes_to_floats(info, self._FLOAT_PRECISION)
                    info = [info[0:3], info[3:6]]

                return info

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def get_fft(self):
        """Get Fast Fourier Transform of vibration data.

        Returns:
            list: A n-elements list, with each element being a list of four
            values: the first is a frequency [Hz] and the other three are the
            corresponding vibration values on the three axis [m/s2].

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                l = IOLinkSensor._SIZE_OF_FDM_LINES
                n = IOLinkSensor._SIZE_OF_FDM * \
                    IOLinkSensor._SIZE_OF_FLOAT_bytes
                while True:
                    info = self._get_measure(IOLinkProtocol.COMMAND_MEAS1_4)

                    if not IOLinkProtocol.BYTES_TRANSMISSION:
                        info = info.split(IOLinkProtocol.TERMINATOR_SEQ)[2:]
                        if len(info) == l:
                            break
                    else:
                        if len(info) == l * n:
                            break

                    print('Missing data. Sending again...')

                if not IOLinkProtocol.BYTES_TRANSMISSION:
                    for i in range(0, len(info)):
                        info[i] = list(
                            map(lambda s: float(s), info[i].split('\t')[0:-1]))
                else:
                    info = [info[i * n:i * n + n] for i in range(0, l)]
                    info = list(map(lambda s: self._bytes_to_floats(s, \
                        self._FLOAT_PRECISION), info))

                return info

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def _set_parameter(self, parameter, value):
        """Execute a 'set' parameter command.

        Args:
            parameter (str): The parameter command to execute.
            value (str): The parameter value to set.

        Returns:
            bool: True if the parameter has been set correctly, False otherwise.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):

                self._master._execute(IOLinkProtocol.COMMAND_ICD,
                    IOLinkProtocol.REQUEST_SLAVE)

                self._master._execute(str(self._position - 1) \
                    + IOLinkProtocol.TERMINATOR_SEQ,
                    IOLinkProtocol.REQUEST_SENSOR_COMMAND)

                self._master._execute(IOLinkProtocol.COMMAND_SET,
                    IOLinkProtocol.REQUEST_PARAMETER_NAME)

                self._master._execute(parameter,
                    IOLinkProtocol.REQUEST_PARAMETER_VALUE)

                self._master._execute(value \
                    + IOLinkProtocol.TERMINATOR_SEQ,
                    IOLinkProtocol.TERMINATOR_SEQ)

                self._master._execute(None, IOLinkProtocol.TERMINATOR_SEQ)

                info = self._master.get_answer()

                self._master._execute(IOLinkProtocol.COMMAND_END,
                    IOLinkProtocol.REQUEST_MOD)

                return True if IOLinkProtocol.MESSAGE_PARAMETER_UPDATED \
                    in info else False

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def set_odr(self, odr):
        """Set accelerometer's output data rate.

        Args:
            odr (:class:`wire_st_sdk.iolink.iolink_protocol.ODR`):
                accelerometer's output data rate.

        Returns:
            bool: True if the parameter has been set correctly, False otherwise.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                return self._set_parameter(
                    IOLinkProtocol.COMMAND_ODR,
                    '{:04d}'.format(odr.value))

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def set_fls(self, fls):
        """Set accelerometer's full scale.

        Args:
            fls (:class:`wire_st_sdk.iolink.iolink_protocol.FLS`):
                accelerometer's full scale.

        Returns:
            bool: True if the parameter has been set correctly, False otherwise.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                return self._set_parameter(
                    IOLinkProtocol.COMMAND_FLS,
                    '{:02d}'.format(fls.value))

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def set_sze(self, sze):
        """Set accelerometer's input array size for FFT.

        Args:
            sze (:class:`wire_st_sdk.iolink.iolink_protocol.SZE`):
                accelerometer's input array size for FFT.

        Returns:
            bool: True if the parameter has been set correctly, False otherwise.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                return self._set_parameter(
                    IOLinkProtocol.COMMAND_SZE,
                    '{:04d}'.format(sze.value))

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def set_sub(self, sub):
        """Set accelerometer's number of subranges.

        Args:
            sub (:class:`wire_st_sdk.iolink.iolink_protocol.SUB`): number of
                subranges.

        Returns:
            bool: True if the parameter has been set correctly, False otherwise.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                return self._set_parameter(
                    IOLinkProtocol.COMMAND_SUB,
                    '{:02d}'.format(sub.value))

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def set_acq(self, acq):
        """Set accelerometer's total acquisition time, which is valid for all
        types of analysis.

        Args:
            acq (int): accelerometer's total acquisition time (must be in the
                range [ACQ_MIN..ACQ_MAX]).

        Returns:
            bool: True if the parameter has been set correctly, False otherwise.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                self._acq = acq
                return self._set_parameter(
                    IOLinkProtocol.COMMAND_ACQ,
                    '{:05d}'.format(acq))

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e

    def set_ovl(self, ovl):
        """Set accelerometer's overlapping percentage between two consecutive
        FFT analysis.

        Args:
            ovl (int): accelerometer's overlapping percentage between two
                consecutive FFT analysis (must be in the range
                [OVL_MIN..OVL_MAX]).

        Returns:
            bool: True if the parameter has been set correctly, False otherwise.

        Raises:
            'SerialException' or 'SerialTimeoutException' are raised if
                something with the serial communication does not work.
            :exc:`wire_st_sdk.utils.wire_st_exceptions.InvalidOperationException`
                is raised if the command has not been executed successfully.
        """
        try:
            with lock_for_object(self._master):
                return self._set_parameter(
                    IOLinkProtocol.COMMAND_OVL,
                    '{:02d}'.format(ovl))

        except (SerialException, SerialTimeoutException,
            InvalidOperationException) as e:
            raise e
